<?php

namespace Drupal\blizz_bulk_creator\Form;

use Drupal\Core\Form\FormStateInterface;
use Drupal\Core\Url;

/**
 * Class BulkcreateUsageFormStep3.
 *
 * Provides the form portion for the third step of
 * creating new bulkcreate usages.
 *
 * @package Drupal\blizz_bulk_creator\Form
 */
class BulkcreateUsageFormStep3 extends ConfigEntityMultistepFormBase {

  /**
   * {@inheritdoc}
   */
  protected function getStepForm(array $form, FormStateInterface $form_state) {

    // Extract the target information of the current settings.
    // TODO - use $this->administrationHelper->getStructuredBulkcreateTargetFieldArray($this)
    $targetstages = array_map(
      function ($item) {
        return explode(':', $item);
      },
      explode('/', $this->store->get('target_field'))
    );

    // Prepare a form element.
    $form['multi_stage'] = [
      '#type' => 'radios',
      '#title' => $this->t('Multi-instantiated field'),
      '#description' => $this->t('Please select the field that will get multi-instantiated for every given value of the bulkcreation (Hint: only fields that can hold more than a single value are presented here for selection).'),
      '#options' => [],
      '#required' => TRUE,
    ];

    // Fetch the field definitions of the current stage.
    $currentfielddefinition = $this->entityHelper->getBundleFields(
      $this->store->get('entity_type_id'),
      $this->store->get('bundle')
    );

    // Inject the options into the prepared field.
    foreach ($targetstages as $delta => $stage) {

      // Determine the cardinality of the current field.
      $cardinality = (int) $stage[1];

      // Only present a selectable option if this
      // field can hold more than a single value.
      if ($cardinality == -1 || $cardinality > 1) {

        // Human-readable shortcut to the field's machine name.
        $field_machine_name = $stage[0];

        // Prepare the field label to present as a selection label.
        $field_label = $currentfielddefinition[$field_machine_name]->getLabel();

        // Insert an option into the prepared field.
        $form['multi_stage']['#options'][$delta] = $field_label;

      }

      // If there are more stages than the current one,
      // prepare the field definition for the next iteration.
      if (!empty($stage[2])) {
        $currentfielddefinition = $this->entityHelper->getBundleFields(
          $stage[2],
          $stage[3]
        );
      }

    }

    return $form;

  }

  /**
   * {@inheritdoc}
   */
  protected function submitStep(array $form, FormStateInterface $form_state) {

    // Shortcut to the configuration entity as well as the type hint.
    /* @var \Drupal\blizz_bulk_creator\Entity\BulkcreateUsage $entity */
    $entity = $this->entity;

    // Get the form input the user filled in.
    $input = $form_state->getUserInput();

    // Load the referenced bulkcreate configuration.
    $bulkcreate_configuration = $this->entityTypeManager->getStorage('bulkcreate_configuration')->load($this->store->get('bulkcreate_configuration'));

    // Get human readable label names.
    $entity_type_label = $this->entityHelper->getEntityTypeOptions()[$this->store->get('entity_type_id')];
    $bundle_label = $this->entityHelper->getEntityBundleOptions($this->store->get('entity_type_id'))[$this->store->get('bundle')];

    // Set the entity values.
    $entity->set('label', "{$entity_type_label}: {$bundle_label}");
    $entity->set('bulkcreate_configuration', $this->store->get('bulkcreate_configuration'));
    $entity->set('entity_type_id', $this->store->get('entity_type_id'));
    $entity->set('bundle', $this->store->get('bundle'));
    $entity->set('target_field', $this->store->get('target_field'));
    $entity->set('multi_stage', (int) $input['multi_stage']);

    // Generate an ID for this entity.
    $entity->set('id', "{$this->store->get('entity_type_id')}-{$this->store->get('bundle')}-{$this->store->get('bulkcreate_configuration')}");

    // Save the entity.
    $entity->save();

    // Invalidate the caches containing base field information.
    $this->cacheTagInvalidator->invalidateTags(['entity_field_info']);

    // Set a message on the frontend.
    drupal_set_message($this->t(
      'Bulkcreations of type %config were enabled on %target_entity.',
      [
        '%config' => $bulkcreate_configuration->label(),
        '%target_entity' => $entity->label(),
      ]
    ));

    // Log a notice to watchdog (or whereever).
    $this->logger->notice(
      'Bulkcreations of type %config were enabled on %target_entity by user %user.',
      [
        '%config' => $bulkcreate_configuration->label(),
        '%target_entity' => $entity->label(),
        '%user' => $this->currentUser()->getAccountName(),
      ]
    );

    // Redirect back to the list view.
    $form_state->setRedirect('blizz_bulk_creator.bulkcreate_usage.list');

  }

  /**
   * {@inheritdoc}
   */
  protected function isFinalStep() {
    return TRUE;
  }

  /**
   * {@inheritdoc}
   */
  protected function getCancelUrl() {
    return new Url('blizz_bulk_creator.bulkcreate_usage.list');
  }

}
